import asyncio
import logging
from asyncio import Task
from dataclasses import dataclass, field
from uuid import uuid4

import witch_doctor
from websockets import WebSocketServerProtocol
from websockets.frames import CloseCode

from xraptor.core.interfaces import Antenna
from xraptor.domain.request import Request
from xraptor.domain.response import Response


@dataclass(slots=True, frozen=True)
class Connection:
    path: str
    connection_hash: int
    remote_ip: str
    ws_server: WebSocketServerProtocol
    connection_id: str
    response_receiver: dict = field(default_factory=dict)

    @classmethod
    def from_ws(cls, ws_server: WebSocketServerProtocol):
        return cls(
            path=ws_server.path,
            connection_hash=ws_server.__hash__(),  # pylint: disable=C2801
            remote_ip=ws_server.remote_address[0],
            ws_server=ws_server,
            connection_id=str(uuid4()),
        )

    def register_response_receiver(self, request: Request):
        self.response_receiver.update(
            {
                request.request_id: asyncio.create_task(
                    self.antenna(request=request)  # pylint: disable=E1120
                ),
            }
        )

    def unregister_response_receiver(self, request: Request):
        if request.request_id in self.response_receiver:
            _task: Task = self.response_receiver[request.request_id]
            _task.cancel()
            del self.response_receiver[request.request_id]

    def _unregister_all(self):
        _r = [*self.response_receiver]
        for request_id in _r:
            if _task := self.response_receiver.get(request_id):
                _task.cancel()
                del self.response_receiver[request_id]

    @witch_doctor.WitchDoctor.injection
    async def antenna(self, request: Request, antenna: Antenna):
        async for data in antenna.subscribe(request.request_id):
            try:
                if isinstance(data, bytes):
                    data = data.decode()
                _response = Response.create(
                    request_id=request.request_id, header={}, payload=data
                )
                await self.ws_server.send(_response.json())
            except Exception as error:  # pylint: disable=W0718
                logging.error(error)

    async def close(self, close_code: CloseCode = CloseCode.NORMAL_CLOSURE):
        self._unregister_all()
        await self.ws_server.close(close_code)



import asyncio
from asyncio import Task
from typing import Self

import witch_doctor

from xraptor.core.interfaces import Antenna


class Broadcast:
    __broadcast_id: str
    __members: list[str]
    __task: Task | None
    __check_task: Task | None
    _broadcasts: dict[str, Self] = {}

    def __init__(self, broadcast_id: str):
        self.__broadcast_id = broadcast_id
        self.__members = []
        self.__task = None
        self.__check_task = None

    @classmethod
    def get(cls, broadcast_id: str) -> Self:
        """
        correct way to get a broadcast instance
        :param broadcast_id: string identifier
        :return: Broadcast object instance
        """
        if cls._broadcasts.get(broadcast_id) is None:
            cls._broadcasts.update({broadcast_id: Broadcast(broadcast_id)})
        return cls._broadcasts.get(broadcast_id)

    @classmethod
    def _delete(cls, broadcast_id: str):
        del cls._broadcasts[broadcast_id]

    def add_member(self, member: str):
        """
        add member on this chat room and if is the first to coming in, will open the room.
        :param member: member is an antenna id coming from request
        :return:
        """
        _members = {*self.__members, member}
        self.__members = list(_members)
        if len(self.__members) == 1:
            self._open()

    def remove_member(self, member: str):
        """
        remove member from this chat room and if is the last to coming out, will close the room.
        :param member: member is an antenna id coming from request
        :return:
        """
        _members = {*self.__members} - {member}
        self.__members = list(_members)
        if len(self.__members) == 0:
            self._close()
            self._delete(self.__broadcast_id)

    def _close(self):
        if self.__task:
            self.__task.cancel()
        if self.__check_task:
            self.__check_task.cancel()

    def _open(self):
        """
        Start to task to listening chat pubsub channel and to check if
        registered members still connected.
        :return:
        """
        self.__task = asyncio.create_task(self._listening())  # pylint: disable=E1120
        self.__check_task = asyncio.create_task(self._check())  # pylint: disable=E1120

    @witch_doctor.WitchDoctor.injection
    async def _check(self, antenna: Antenna, frequency: int = 7):
        """
        check each 7 seconds, using PUBSUB NUMSUB redis command, if each
        member on this room still connected, otherwise the member will be removed.
        :return:
        """
        while True:
            _to_check = [*self.__members]
            _status = await asyncio.gather(*[antenna.is_alive(i) for i in _to_check])
            _ = [
                None if is_alive else self.remove_member(_to_check[index])
                for index, is_alive in enumerate(_status)
            ]
            await asyncio.sleep(frequency)

    @witch_doctor.WitchDoctor.injection
    async def _listening(self, antenna: Antenna):
        """
        start listening each message from the chat room pubsub
        channel and broadcast it to each member in this room.
        :return:
        """
        async for data in antenna.subscribe(self.__broadcast_id):
            await asyncio.gather(*[antenna.post(i, data) for i in self.__members])



from abc import ABC, abstractmethod
from typing import AsyncIterator, Awaitable


class Antenna(ABC):
    @abstractmethod
    async def subscribe(self, antenna_id: str) -> AsyncIterator[str]:
        """
        async generator that will yield message from the key's channel
        :param antenna_id: pubsub channel
        :return: str message async generator
        """

    @abstractmethod
    async def post(self, antenna_id: str, message: str) -> Awaitable:
        """
        async function that will publish a message to a key's channel
        :param antenna_id: pubsub channel
        :param message: message
        :return:
        """

    @abstractmethod
    async def is_alive(self, antenna_id: str) -> Awaitable[bool]:
        """
        verify that antenna_id still alive
        :param antenna_id:
        :return:
        """

    @classmethod
    @abstractmethod
    def set_config(cls, config: dict):
        """
        set config map for this antenna
        :param config:
        :return:
        """



from . import antenna_implementations as antennas
from .broadcaster import Broadcast
from .domain.request import Request
from .domain.response import Response
from .server import XRaptor

__all__ = ["XRaptor", "antennas", "Request", "Response", "Broadcast"]
